@@ -1,49 +1,54 @@
 """Classes for odorants, mixtures, chemical orders, etc."""
 
 import base64
-from collections import OrderedDict
 import io
 import json
-from datetime import datetime
-from urllib.request import urlopen
-from urllib.error import HTTPError
-from urllib.parse import quote
 import re
 import time
+import warnings
+from collections import OrderedDict
+from datetime import datetime
+from urllib.error import HTTPError
+from urllib.parse import quote
+from urllib.request import urlopen
 
-from IPython.display import display, Image
 import numpy as np
 import pubchempy as pcp
+from IPython.display import display
 from PIL import Image
+
 import quantities as pq
+from pyrfume import load_data, logger, tqdm, trange
+from pyrfume.physics import mackay
 from quantities.constants.statisticalmechanics import R
-import warnings
 
 try:
     from rdkit import Chem
-    from rdkit.Chem import Draw
+    from rdkit.Chem import Draw, AllChem, SaltRemover
 except ImportError:
-    warnings.warn("Parts of rdkit could not be imported; try installing rdkit via conda",
-                  UserWarning)
-
-from pyrfume import load_data, logger, tqdm, trange
-from pyrfume.physics import mackay
+    warnings.warn(
+        "Parts of mordred and/or rdkit could not be imported; try installing rdkit via conda",
+        UserWarning,
+    )
 
 ROOM_TEMP = (22 + 273.15) * pq.Kelvin
 ROOM_PRESSURE = 1 * pq.atm
 GAS_MOLAR_DENSITY = ROOM_PRESSURE / (R * ROOM_TEMP)
 
-ODORANTS_BASIC_INFO_PATH = 'odorants/all-cids-properties.csv'
-ODORANT_SOURCES_PATH = 'odorants/all-cids.csv'
+ODORANTS_BASIC_INFO_PATH = "odorants/all-cids-properties.csv"
+ODORANT_SOURCES_PATH = "odorants/all-cids.csv"
 
 
 class Solution:
+    components: dict = {}
+    
     def __init__(self, components: dict, date_created: datetime.timestamp = None):
         self.total_volume = 0 * pq.mL
         assert isinstance(components, dict), "Components must be a dict"
         for component, volume in components.items():
-            assert isinstance(component, (Compound, Solution)), \
-                "Each component must be a Compound or a Solution"
+            assert isinstance(
+                component, (Compound, Solution)
+            ), "Each component must be a Compound or a Solution"
             try:
                 volume = volume.rescale(pq.mL)
             except ValueError:
@@ -73,52 +78,51 @@ class Solution:
     def molecules(self):
         """Returns a dictionary with the moles of each Molecule"""
         compounds = self.compounds
-        assert all([c.density for c, v in compounds.items() if v
-                    and not c.is_solvent]), \
-            ("All non-solvent compounds must have a known density "
-             "in order to compute moles")
-        assert all([c.molecular_weight for c, v in compounds.items() if v
-                    and not c.is_solvent]), \
-            ("All non-solvent compounds must have a known molecular weight "
-             "in order to compute moles")
-        return {c.molecule: (v * c.molarity).rescale(pq.mol)
-                for c, v in self.compounds.items() if v}
+        assert all([c.density for c, v in compounds.items() if v and not c.is_solvent]), (
+            "All non-solvent compounds must have a known density " "in order to compute moles"
+        )
+        assert all([c.molecular_weight for c, v in compounds.items() if v and not c.is_solvent]), (
+            "All non-solvent compounds must have a known molecular weight "
+            "in order to compute moles"
+        )
+        return {
+            c.molecule: (v * c.molarity).rescale(pq.mol) for c, v in self.compounds.items() if v
+        }
 
     @property
     def molarities(self):
         """Returns a dictionary with the molarity of each Molecule"""
-        return {m: mol/self.total_volume
-                for m, mol in self.molecules.items() if mol}
+        return {m: mol / self.total_volume for m, mol in self.molecules.items() if mol}
 
     @property
     def mole_fractions(self):
         """Returns a dictionary with the mole fraction of each Molecule"""
         molecules = self.molecules
-        assert([moles for molecule, moles in molecules.items()]), \
-            ("All compounds must have a known number of moles "
-             "in order to compute mole fraction")
+        assert [moles for molecule, moles in molecules.items()], (
+            "All compounds must have a known number of moles " "in order to compute mole fraction"
+        )
         # A Quantities bug prevents me from simply summing molecules.values()
         total_moles = 0 * pq.mol
         for moles in molecules.values():
             total_moles += moles
-        return {molecule: moles/total_moles
-                for molecule, moles in molecules.items()}
+        return {molecule: moles / total_moles for molecule, moles in molecules.items()}
 
     def mole_fraction(self, molecule):
-        return self.mole_fractions[molecule] \
-               if molecule in self.mole_fractions else 0
+        return self.mole_fractions[molecule] if molecule in self.mole_fractions else 0
 
     @property
     def dilutions(self):
-        return {c.molecule: self.total_volume/c.volume
-                for c in self.compounds if not c.is_solvent}
+        return {
+            c.molecule: self.total_volume / c.volume for c in self.compounds if not c.is_solvent
+        }
 
     @property
     def partial_pressures(self):
         """Computes partial pressures for each odorant
         in the mixture using Raoult's law"""
-        return {m: self.mole_fraction(m)*m.vapor_pressure
-                for m in self.molecules if m.vapor_pressure}
+        return {
+            m: self.mole_fraction(m) * m.vapor_pressure for m in self.molecules if m.vapor_pressure
+        }
 
     def partial_pressure(self, molecule):
         return self.partial_pressures[molecule]
@@ -126,8 +130,9 @@ class Solution:
     @property
     def total_pressure(self):
         """Computes total pressure of the vapor using Dalton's law"""
-        partial_pressures = self.partial_pressures.values()
-        return sum(partial_pressures)
+        preferred_units = pq.Pa
+        partial_pressures = [pressure.rescale(preferred_units) for pressure in self.partial_pressures.values()]
+        return preferred_units * np.sum(partial_pressures)
 
     @property
     def vapor_concentrations(self):
@@ -136,7 +141,7 @@ class Solution:
         pp = self.partial_pressures
         result = {}
         for m, p in pp.items():
-            ratio = (p/pq.atm).simplified
+            ratio = (p / pq.atm).simplified
             assert ratio.units == pq.dimensionless
             result[m] = float(ratio)
         return result
@@ -147,71 +152,18 @@ class Solution:
     @property
     def molar_evaporation_rates(self):
         mf = self.mole_fractions
-        result = {molecule: mole_fraction*molecule.molar_evaporation_rate
-                  for molecule, mole_fraction in mf.items()}
+        result = {
+            molecule: mole_fraction * molecule.molar_evaporation_rate
+            for molecule, mole_fraction in mf.items()
+        }
         return result
 
-
-class Compound:
-    def __init__(self, chemical_order, stock='',
-                 date_arrived=None, date_opened=None, is_solvent=False):
-        self.chemical_order = chemical_order
-        self.date_arrived = date_arrived if date_arrived else datetime.now
-        self.date_opened = date_opened
-        self.is_solvent = is_solvent
-
-    # ChemicalOrder
-    chemical_order = None
-    # Stock number (supplied by vendor, usually on bottle)
-    stock = ''
-    # Date arrived at the lab/clinic
-    date_arrived = None
-    # Date opened
-    date_opened = None
-    # Is it a solvent?
-    is_solvent = False
-
-    def __getattr__(self, attr):
-        """If no attribute is found, try looking up on the
-        ChemicalOrder or the Molecule"""
-        try:
-            return getattr(self.chemical_order, attr)
-        except AttributeError:
-            return getattr(self.chemical_order.molecule, attr)
-
-
-class ChemicalOrder:
-    def __init__(self, molecule, vendor, part_id,
-                 purity=1, known_impurities=None):
-        self.molecule = molecule
-        self.vendor = vendor
-        self.part_id = part_id
-        self.purity = purity
-        self.known_impurities = known_impurities
-
-    # Molecule
-    molecule : 'Molecule' = None
-    # Vendor, e.g. Sigma-Aldrich
-    vendor : 'Vendor' = None
-    # ID number of compound at vendor
-    part_id : str = ''
-    # Reported purity as a fraction
-    purity : float = 1
-    # List of known impurities (Molecules)
-    known_impurities : list = None
-
-
-class Vendor:
-    def __init__(self, name, url):
-        self.name = name
-        self.url = url
-
-    name : str = ''
-    url : str = ''
+    def molar_evaporation_rate(self, molecule):
+        return self.molar_evaporation_rates[molecule]
 
 
 class Molecule:
-    def __init__(self, cid, name=None, fill=False):
+    def __init__(self, cid: int, name: str=None, fill: bool=False):
         self.cid = cid
         if fill:
             self.fill_details()
@@ -219,25 +171,25 @@ class Molecule:
             self.name = name
 
     # Integer Chemical ID number (CID) from PubChem
-    cid : int = 0
+    cid: int = 0
     # Chemical Abstract Service (CAS) number
-    cas : str = ''
+    cas: str = ""
     # Principal name
-    name : str = ''
+    name: str = ""
     # Synonyms
-    synonyms : str = ''
+    synonyms: str = ""
     # IUPAC name (long, unique name)
-    iupac : str = ''
+    iupac: str = ""
     # Density (pq.g / pq.ml)
-    density : float = None
+    density: float = None
     # Vapor pressure (pq.Pa)
-    vapor_pressure : float = None
+    vapor_pressure: float = None
     # Molecular weight (pq.g / pq.mol)
-    molecular_weight : float = None
+    molecular_weight: float = None
 
     @property
     def molarity(self):
-        if not self.molecular_weight:
+        if not (self.molecular_weight and self.density):
             result = None
         else:
             result = self.density / self.molecular_weight
@@ -250,24 +202,23 @@ class Molecule:
 
     def fill_details(self):
         assert self.cid is not None
-        url_template = ("https://pubchem.ncbi.nlm.nih.gov/"
-                        "rest/pug/compound/cid/%d/property/"
-                        "%s/JSON")
-        property_list = ['MolecularWeight', 'IsomericSMILES']
-        url = url_template % (self.cid, ','.join(property_list))
+        url_template = (
+            "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/cid/%d/property/" "%s/JSON"
+        )
+        property_list = ["MolecularWeight", "IsomericSMILES"]
+        url = url_template % (self.cid, ",".join(property_list))
         json_data = url_to_json(url)
-        details = json_data['PropertyTable']['Properties'][0]
+        details = json_data["PropertyTable"]["Properties"][0]
 
         def convert(name):
-            s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
-            return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
+            s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
+            return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()
 
         for key, value in details.items():
-            if key == 'CID':
-                assert value == self.cid, \
-                    "REST API CID does not match provided CID"
+            if key == "CID":
+                assert value == self.cid, "REST API CID does not match provided CID"
             key = convert(key)
-            if key == 'molecular_weight':
+            if key == "molecular_weight":
                 value *= pq.g / pq.mol
             setattr(self, key, value)
 
@@ -275,28 +226,25 @@ class Molecule:
             self.name = self.get_name_from_api()
 
     def get_name_from_api(self):
-        url_template = ("https://pubchem.ncbi.nlm.nih.gov/"
-                        "rest/pug/compound/cid/%d/synonyms/JSON")
+        url_template = "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/cid/%d/synonyms/JSON"
         url = url_template % (self.cid)
         json_data = url_to_json(url)
         name = None
         if json_data:
-            information = json_data['InformationList']['Information'][0]
-            synonyms = information['Synonym']
+            information = json_data["InformationList"]["Information"][0]
+            synonyms = information["Synonym"]
             name = synonyms[0].lower()
         return name
 
     def get_cid_from_api(self):
-        url_template = ("https://pubchem.ncbi.nlm.nih.gov/"
-                        "rest/pug/compound/%s/%s/cids/JSON")
-        options = [getattr(self, x) for x in ('cas', 'name')
-                   if len(getattr(self, x))]
+        url_template = "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/%s/%s/cids/JSON"
+        options = [getattr(self, x) for x in ("cas", "name") if len(getattr(self, x))]
         cid = None
         query = self.name
         for option in options:
             url = url_template % (option, query)
             json_data = url_to_json(url)
-        cid = json_data['IdentifierList']['CID'][0]
+        cid = json_data["IdentifierList"]["CID"][0]
         return cid
 
     def __eq__(self, other):
@@ -304,7 +252,7 @@ class Molecule:
             return self.cid == other.cid
         else:
             return self.name == self.name
- 
+
     def __lt__(self, other):
         if self.cid:
             return self.cid < other.cid
@@ -316,22 +264,81 @@ class Molecule:
 
     def __repr__(self):
         if self.cid and self.name:
-            result = '%d (%s)' % (self.cid, self.name)
+            result = "%d (%s)" % (self.cid, self.name)
         elif self.cid:
-            result = '%d' % self.cid
+            result = "%d" % self.cid
         elif self.name:
-            result = '%s' % self.name
+            result = "%s" % self.name
         else:
-            result = 'Unknown'
+            result = "Unknown"
         return result
 
 
+class Vendor:
+    def __init__(self, name: str, url: str):
+        self.name = name
+        self.url = url
+
+    name: str = ""
+    url: str = ""
+
+
+class ChemicalOrder:
+    def __init__(self, molecule: "Molecule", vendor: "Vendor", part_id: str, purity: float=1, known_impurities: list=None):
+        self.molecule = molecule
+        self.vendor = vendor
+        self.part_id = part_id
+        self.purity = purity
+        self.known_impurities = known_impurities
+
+    # Molecule
+    molecule: Molecule = None
+    # Vendor, e.g. Sigma-Aldrich
+    vendor: Vendor = None
+    # ID number of compound at vendor
+    part_id: str = ""
+    # Reported purity as a fraction
+    purity: float = 1
+    # List of known impurities (Molecules)
+    known_impurities: list = None
+
+
+class Compound:
+    def __init__(
+        self, chemical_order: ChemicalOrder, stock: str="", date_arrived: datetime=None, date_opened: datetime=None, is_solvent: bool=False
+    ):
+        self.chemical_order = chemical_order
+        self.stock = stock
+        self.date_arrived = date_arrived if date_arrived else datetime.now
+        self.date_opened = date_opened
+        self.is_solvent = is_solvent
+
+    # ChemicalOrder
+    chemical_order: ChemicalOrder = None
+    # Stock number (supplied by vendor, usually on bottle)
+    stock: str = ""
+    # Date arrived at the lab/clinic
+    date_arrived: datetime = None
+    # Date opened
+    date_opened: datetime = None
+    # Is it a solvent?
+    is_solvent: bool = False
+
+    def __getattr__(self, attr):
+        """If no attribute is found, try looking up on the
+        ChemicalOrder or the Molecule"""
+        try:
+            return getattr(self.chemical_order, attr)
+        except AttributeError:
+            return getattr(self.chemical_order.molecule, attr)
+
+
 def url_to_json(url, verbose=True) -> str:
     json_data = None
     msgs = []
     try:
         page = urlopen(url)
-        string = page.read().decode('utf-8')
+        string = page.read().decode("utf-8")
         json_data = json.loads(string)
     except HTTPError:
         msgs.append("HTTPError for query '%s'" % url)
@@ -341,20 +348,34 @@ def url_to_json(url, verbose=True) -> str:
     return json_data
 
 
-def canonical_smiles(smiles: str, kekulize: bool = False) -> str:
-    """Use rdkit to convert the `smiles` string to canonical form"""
+def deisomerize_smiles(smiles: str) -> str:
     mol = Chem.MolFromSmiles(smiles)
-    if mol: # If a mol object was successfully create (i.e. not `None`)
-        if kekulize:
-            Chem.Kekulize(mol)
-        smiles = Chem.MolToSmiles(mol, isomericSmiles=True)
-    else: # No mol object means the `smiles` string was invalid
-        smiles = ''
+    if mol:  # If a mol object was successfully create (i.e. not `None`)
+        smiles = Chem.MolToSmiles(mol, isomericSmiles=False)
+    else:
+        smiles = smiles.replace('@','').replace('@@','').replace('/','').replace('\\','')
     return smiles
 
 
-def get_cids(identifiers: list, kind: str ='name', verbose: bool =True,
-             wait: float = 0, results: dict = None) -> dict:
+def canonical_smiles(smiles: str, kekulize: bool = False) -> str:
+    """Use rdkit to convert the `smiles` string to canonical form"""
+    mol = Chem.MolFromSmiles(smiles)
+    if mol:  # If a mol object was successfully create (i.e. not `None`)
+        if kekulize:
+            Chem.Kekulize(mol)
+        smiles = Chem.MolToSmiles(mol, isomericSmiles=True)
+    else:  # No mol object means the `smiles` string was invalid
+        smiles = ""
+    return smiles
+
+
+def get_cids(
+    identifiers: list,
+    kind: str = "name",
+    verbose: bool = True,
+    wait: float = 0,
+    results: dict = None,
+) -> dict:
     """Return CIDs for molecule based on any synonym,
     including a chemical name or a CAS"""
     kind = kind.lower()
@@ -362,7 +383,7 @@ def get_cids(identifiers: list, kind: str ='name', verbose: bool =True,
         results = {}
     p = tqdm(identifiers)
     for identifier in p:
-        p.set_description(identifier)
+        p.set_description(str(identifier))
         cid = get_cid(identifier, kind=kind, verbose=verbose)
         if not cid:
             logger.warning("Could not find %s" % identifier)
@@ -372,11 +393,14 @@ def get_cids(identifiers: list, kind: str ='name', verbose: bool =True,
     return results
 
 
-def get_cid(identifier: str, kind: str = 'name', verbose: bool = True,
-            fix_smiles_on_error: bool = True) -> int:
+def get_cid(
+    identifier: str, kind: str = "name", verbose: bool = True, fix_smiles_on_error: bool = True
+) -> int:
     """Return data about a molecule from any synonym,
     including a chemical name or a CAS"""
-    kind=kind.lower()
+    kind = kind.lower()
+    if isinstance(identifier, float) and np.isnan(identifier):
+        return 0
     try:
         result = pcp.get_cids(identifier, namespace=kind)
     except pcp.BadRequestError:
@@ -386,10 +410,9 @@ def get_cid(identifier: str, kind: str = 'name', verbose: bool = True,
         cid = 0
     else:
         if (len(result) > 1) and verbose:
-            logger.warning("Multiple CIDs for %s: %s" %
-                       (identifier, result))
+            logger.warning("Multiple CIDs for %s: %s" % (identifier, result))
         cid = result[0]
-    if not cid and kind == 'smiles' and fix_smiles_on_error:
+    if not cid and kind == "smiles" and fix_smiles_on_error:
         # Retry with canonical SMILES
         identifier = canonical_smiles(identifier)
         if identifier:
@@ -399,44 +422,47 @@ def get_cid(identifier: str, kind: str = 'name', verbose: bool = True,
 
 def from_cids(cids: list, property_list: bool = None) -> list:
     if property_list is None:
-        property_list = ['MolecularWeight', 'IsomericSMILES', 'IUPACName']
+        property_list = ["MolecularWeight", "IsomericSMILES", "IUPACName"]
     result = []
     chunk_size = 100
     for start in trange(0, len(cids), chunk_size):
         stop = min(start + chunk_size, len(cids))
-        msg = "Retrieving %d through %d" % (start, stop-1)
-        cid_subset = [str(x) for x in [cid for cid in cids[start:stop] if int(cid)>0 and cid is not None]]
-        cid_subset = ','.join(cid_subset)
-        properties_template = ("https://pubchem.ncbi.nlm.nih.gov/"
-                               "rest/pug/compound/cid/%s/property/"
-                               "%s/JSON")
-        url = properties_template % (cid_subset, ','.join(property_list))
+        print("Retrieving %d through %d" % (start, stop - 1))
+        cid_subset = [
+            str(x) for x in [cid for cid in cids[start:stop] if int(cid) > 0 and cid is not None]
+        ]
+        cid_subset = ",".join(cid_subset)
+        properties_template = (
+            "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/cid/%s/property/" "%s/JSON"
+        )
+        url = properties_template % (cid_subset, ",".join(property_list))
         json_data = url_to_json(url)
-        data = json_data['PropertyTable']['Properties']
+        data = json_data["PropertyTable"]["Properties"]
 
-        synonyms_template = ("https://pubchem.ncbi.nlm.nih.gov/"
-                             "rest/pug/compound/cid/%s/synonyms/JSON")
+        synonyms_template = (
+            "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/cid/%s/synonyms/JSON"
+        )
         url = synonyms_template % (cid_subset)
         json_data = url_to_json(url)
-        information = json_data['InformationList']['Information']
+        information = json_data["InformationList"]["Information"]
         for i, d in enumerate(data):
             try:
-                synonyms = information[i]['Synonym']
+                synonyms = information[i]["Synonym"]
             except KeyError:
                 try:
-                    d['name'] = d['IUPACName']
+                    d["name"] = d["IUPACName"]
                 except KeyError:
-                    d['name'] = ''
+                    d["name"] = ""
             else:
-                d['name'] = synonyms[0].lower()
+                d["name"] = synonyms[0].lower()
         result += data
     return result
 
 
 def cids_to_smiles(cids: list) -> dict:
     """Returns an ordered dictionary of SMILES strings with CIDs as keys"""
-    info = from_cids(cids, property_list=['IsomericSMILES'])
-    smiles = {item['CID']: item['IsomericSMILES'] for item in info}
+    info = from_cids(cids, property_list=["IsomericSMILES"])
+    smiles = {item["CID"]: item["IsomericSMILES"] for item in info}
     return smiles
 
 
@@ -445,18 +471,19 @@ def cids_to_cas(cids: list) -> OrderedDict:
     chunk_size = 100
     for start in trange(0, len(cids), chunk_size):
         stop = min(start + chunk_size, len(cids))
-        #msg = "Retrieving %d through %d" % (start, stop-1)
+        # msg = "Retrieving %d through %d" % (start, stop-1)
         cid_subset = [str(x) for x in cids[start:stop]]
-        cid_subset = ','.join(cid_subset)
-        synonyms_template = ("https://pubchem.ncbi.nlm.nih.gov/"
-                             "rest/pug/compound/cid/%s/synonyms/JSON")
+        cid_subset = ",".join(cid_subset)
+        synonyms_template = (
+            "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/cid/%s/synonyms/JSON"
+        )
         url = synonyms_template % (cid_subset)
         json_data = url_to_json(url)
-        information = json_data['InformationList']['Information']
+        information = json_data["InformationList"]["Information"]
         for i, info in enumerate(information):
-            cid = info['CID']
+            cid = info["CID"]
             try:
-                synonyms = info['Synonym']
+                synonyms = info["Synonym"]
             except KeyError:
                 result[cid] = []
             else:
@@ -468,21 +495,23 @@ def cids_to_cas(cids: list) -> OrderedDict:
 def cas_from_synonyms(synonyms: list) -> list:
     result = []
     for s in synonyms:
-        if re.match('^[0-9]+\-[0-9]+\-[0-9]+$', s):
+        if re.match(r"^[0-9]+\-[0-9]+\-[0-9]+$", s):
             result.append(s)
     return result
 
-            
-def cactus(identifier: str, output: str = 'cas') -> str:
+
+def cactus(identifier: str, output: str = "cas") -> str:
     url_template = "https://cactus.nci.nih.gov/chemical/structure/%s/%s"
+    identifier = identifier.replace(' ', '%20')
     url = url_template % (identifier, output)
     page = urlopen(url)
-    result = page.read().decode('utf-8')
+    result = page.read().decode("utf-8")
     return result
 
 
 def cactus_image(smiles: str) -> None:
     url_template = "https://cactus.nci.nih.gov/chemical/structure/%s/image"
+    smiles = smiles.replace(' ', '%20')
     url = url_template % smiles
     image_data = urlopen(url).read()
     image = Image(image_data)
@@ -492,8 +521,9 @@ def cactus_image(smiles: str) -> None:
 def get_compound_summary(cid: int, heading: str):
     """Get summary info about `heading` from PubChem for the compound
     given by `cid`.  Example heading: 'Physical Description'"""
-    url_template = ("https://pubchem.ncbi.nlm.nih.gov/"
-                    "rest/pug_view/data/compound/%d/JSON?heading=%s")
+    url_template = (
+        "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug_view/data/compound/%d/JSON?heading=%s"
+    )
     escaped_heading = quote(heading)  # Escape the string
     url = url_template % (cid, escaped_heading)
     json_data = url_to_json(url, verbose=False)
@@ -502,7 +532,7 @@ def get_compound_summary(cid: int, heading: str):
 
 def get_compound_odor(cid, raw=False):
     info = []
-    for heading in ['Odor', 'Physical Description']:
+    for heading in ["Odor", "Physical Description"]:
         json_data = get_compound_summary(cid, heading)
         if raw:
             info += [] if json_data is None else [json_data]
@@ -516,9 +546,9 @@ def _parse_odor_info(info, odors=None, any_string=False):
         odors = []
     if isinstance(info, dict):
         for key, value in info.items():
-            if key == 'TOCHeading' and value == 'Odor':
+            if key == "TOCHeading" and value == "Odor":
                 any_string = True
-            if key == 'String' and (any_string or 'odor' in value.lower()):
+            if key == "String" and (any_string or "odor" in value.lower()):
                 odors.append(value)
             else:
                 _parse_odor_info(value, odors=odors, any_string=any_string)
@@ -533,9 +563,9 @@ def _parse_other_info(info, records=None):
         records = []
     if isinstance(info, dict):
         for key, value in info.items():
-            if key == 'String':
+            if key == "String":
                 records.append(value)
-            elif key == 'Value' and 'Number' in value:
+            elif key == "Value" and "Number" in value:
                 records.append(value)
             else:
                 _parse_other_info(value, records=records)
@@ -556,14 +586,42 @@ def smiles_to_image(smiles, png=True, b64=False, crop=True, padding=10, size=300
     if crop:
         image = crop_image(image, padding=padding)
     if png:
-        image.save(buffer, format='PNG')
+        image.save(buffer, format="PNG")
         image = buffer.getvalue()
     if b64:
         assert png, "Can only base64 encode PNG data, not a PIL image"
-        image = base64.b64encode(image).decode('utf8')
+        image = base64.b64encode(image).decode("utf8")
     return image
 
 
+def smiles_to_mol(smiles: list, max_attempts: int=10, use_random_coords: bool=False, deisomerize=False) -> dict:
+    if deisomerize:
+        f = deisomerize_smiles
+    else:
+        f = lambda x: x
+    mols_raw = [Chem.MolFromSmiles(f(smi)) for smi in smiles]
+    logger.info("Computing 3D coordinates...")
+    s = SaltRemover.SaltRemover()
+    mols = {}
+    n = len(mols_raw)
+    pbar = tqdm(total=n)
+    for i, mol in enumerate(mols_raw):
+        pbar.update()
+        logger.debug("Embedding %s" % smiles[i])
+        try:
+            mol = s.StripMol(mol, dontRemoveEverything=True)
+            mol = Chem.AddHs(mol)
+            AllChem.Compute2DCoords(mol)
+            AllChem.EmbedMolecule(mol, maxAttempts=max_attempts, useRandomCoords=use_random_coords)
+            AllChem.UFFOptimizeMolecule(mol)  # Is this deterministic?
+        except Exception as e:
+            logger.warning("Exception for %s: %s" % (smiles[i], str(e)))
+        else:
+            mols[smiles[i]] = mol
+    logger.info("Finished embedding all molecules")
+    return mols
+
+
 def crop_image(img, padding=0):
     """Crop white out of a PIL image."""
     as_array = np.array(img)  # N x N x (r,g,b,a)
@@ -573,9 +631,8 @@ def crop_image(img, padding=0):
     xs, ys = np.nonzero(has_content)
     x_range = max([min(xs) - padding, 0]), min([max(xs) + padding, as_array.shape[0]])
     y_range = max([min(ys) - padding, 0]), min([max(ys) + padding, as_array.shape[1]])
-    as_array_cropped = as_array[
-        x_range[0]:x_range[1], y_range[0]:y_range[1], 0:3]
-    img = Image.fromarray(as_array_cropped, mode='RGB')
+    as_array_cropped = as_array[x_range[0] : x_range[1], y_range[0] : y_range[1], 0:3]
+    img = Image.fromarray(as_array_cropped, mode="RGB")
     return img
 
 
@@ -604,9 +661,9 @@ def all_smiles():
     """All SMILES found in the file at ODORANTS_BASIC_INFO_PATH.
     May contain duplicates (if two CIDs give the same SMILES)"""
     df = all_odorants()
-    return list(df['SMILES'])
+    return list(df["SMILES"])
 
 
-if __name__ == '__main__':
+if __name__ == "__main__":
     x = Molecule(325, fill=True)
     print(x.__dict__)
